// JS target: IE 11+

// polyfills:
if (!Element.prototype.matches) {
  Element.prototype.matches =
    Element.prototype.msMatchesSelector ||
    Element.prototype.webkitMatchesSelector;
}
if (!Element.prototype.closest) {
  Element.prototype.closest = function(s) {
    var el = this;
    if (!document.documentElement.contains(el)) return null;
    do {
      if (el.matches(s)) return el;
      el = el.parentElement || el.parentNode;
    } while (el !== null && el.nodeType === 1);
    return null;
  };
}


function plainArray(arrayLike) {
  return Array.prototype.slice.apply(arrayLike);
}

function each(arr, cb) {
  Array.prototype.forEach.call(arr, cb);
}

function find(arr, fn) {
  for (let i = 0, len = arr.length; i < len; i++) {
    if (fn(arr[i], i)) return arr[i];
  }
}

function isHidden(el) {
  return el.offsetParent === null || /* ie 9-10 */ el.parentNode === document.body;
}

function pick(obj, keys) {
  const out = {};
  keys.forEach(key => {
    if (key in obj) {
      out[key] = obj[key];
    }
  });
  return out;
}

function assign(to, varArgs) {
  for (let index = 1; index < arguments.length; index++) {
    const nextSource = arguments[index];
    if (nextSource != null) {
      for (const nextKey in nextSource) {
        if (Object.prototype.hasOwnProperty.call(nextSource, nextKey)) {
          to[nextKey] = nextSource[nextKey];
        }
      }
    }
  }
  return to;
}

function toggleClass(el, className, value) {
  if (value) {
    el.classList.add(className);
  } else {
    el.classList.remove(className);
  }
}

// measure the height of each element within the views and place them in approximately equal columns
function columnize(el, columns) {
  const children = Array.prototype.slice.apply(el.children);
  const colHeight = children.reduce((a, n) => a + n.offsetHeight, 0) / columns;

  const columnsData = [];
  for (let i = 0; i < columns; i++) {
    columnsData.push({ height: 0, elements: [] });
  }

  let column = 0;
  children.forEach(child => {
    const colData = columnsData[column];
    colData.elements.push(child);
    colData.height += child.offsetHeight;
    if (colData.height >= colHeight) {
      column++;
    }
  });

  const rowEl = document.createElement('row');
  rowEl.className = 'row';
  columnsData.forEach(colData => {
    const colEl = document.createElement('div');
    colEl.className = 'col-lg col-lg--1';
    colData.elements.forEach(child => colEl.appendChild(child));
    rowEl.appendChild(colEl);
  });
  el.appendChild(rowEl);
}

class View {
  constructor() {
    this.el = null;
  }

  static attach(el) {
    const instance = new this();
    instance.attach(el);
  }

	bindEvents(ev) {
		const self = this;
		for (const key in ev) {
			const components = key.split(' ');
			const event = components[0], filter = components.slice(1).join(' ');
			const callback = ev[key];
			const el = this.el;
			if (event && callback) {
				if (filter) {
					el.addEventListener(event, e => {
						const filteredEl = e.target.closest(filter);
						if (filteredEl) {
							callback.call(self, filteredEl, e);
						}
					});
				} else {
					el.addEventListener(event, e => {
						callback.call(self, e.target, e);
					});
				}
			}
		}
	}
}

class TagSummary extends View {
  constructor() {
    super();
    this.tagsEl = null;
    this.showAllEl = null;
    this.collapsed = false;
  }

  attach(el) {
    this.el = el;
    this.tagsEl = el.querySelector('.tag-summary__tags');
    this.showAllEl = el.querySelector('.tag-summary__show-all');
    this.checkCollapsable();
  }

  checkCollapsable() {
    this.setCollapsed(false);
    const collapsable = this.tagsEl.offsetHeight > 50;
    if (collapsable) {
      this.el.querySelector('.tag-summary__show-all').classList.remove('d-none');
      this.showAllEl.addEventListener('click', () => this.setCollapsed(!this.collapsed));
      this.setCollapsed(true);
    }
  }

  setCollapsed(collapsed) {
    this.collapsed = collapsed;
    toggleClass(this.el, 'tag-summary--collapsed', collapsed);
    toggleClass(this.showAllEl.querySelector('.caret'), 'caret--up', !collapsed);
  }
}

class Dropdown extends View {
  constructor() {
    super();
    this.visible = false;
    this.handleToggleClick = this.handleToggleClick.bind(this);
    this.cancel = this.cancel.bind(this);
  }

  attach(el) {
    this.el = el;
    const toggle = this.el.querySelector('.dropdown__toggle')
    if (toggle) toggle.addEventListener('click', this.handleToggleClick);
  }

  handleToggleClick(e) {
    e.preventDefault();
    this.setVisible(!this.visible);
  }

  cancel() {
    this.setVisible(false);
  }

  setVisible(visible) {
    if (this.visible == visible) return;
    this.visible = visible;
    setTimeout(() => {
      if (visible) {
        document.body.addEventListener('click', this.cancel);
      } else {
        document.body.removeEventListener('click', this.cancel);
      }
    });
    toggleClass(this.el, 'dropdown--open', visible);
  }
}

class Gallery extends View {
  constructor() {
    super();
    this.r18 = false;
    this.toggle = null;
    this.handleSwitchClick = this.handleSwitchClick.bind(this);
  }

  attach(el) {
    this.el = el;
    this.toggle = this.el.querySelector('.gallery-r18-toggle');
    if (this.toggle) this.toggle.addEventListener('click', this.handleSwitchClick);
    this.bindEvents({
      'click .gallery__image-link': this.imageClick
    });
  }

  handleSwitchClick() {
    this.setR18(!this.r18);
  }

  setR18(r18) {
    this.r18 = r18;
    toggleClass(this.el, 'gallery--show-r18', r18);
    toggleClass(this.toggle, 'switch--on', r18);
  }

  imageClick(el, e) {
    e.preventDefault();
    this.openLightbox({ currentEl: el });
  }

  openLightbox({ currentEl }) {
    const images = [];
    let currentIndex = 0;
    each(this.el.querySelectorAll('.gallery__image-link'), el => {
      if (el.classList.contains('gallery__image--r18') && !this.r18) return;
      const imageData = this.extractImageData(el);
      images.push(imageData);
      if (el == currentEl) currentIndex = images.length - 1;
    });

    if (images.length == 0) return;

    const lightbox = new Lightbox({ images, index: currentIndex });
    lightbox.show();
  }

  extractImageData(el) {
    const size = el.dataset.fullImageSize.split('x').map(x => Number(x));
    return {
      src: el.href,
      thumbSrc: el.querySelector('.gallery__image').src,
      width: size[0],
      height: size[1]
    };
  }
}

class Lightbox extends View {
  constructor({ images, index }) {
    super();
    this.images = images.map((x, i) => ({ ...x, index: i }));
    this.index = index || 0;
    this.currentPreload = false;
    this.preloadQueue = [];
    this.handleKeyDown = this.handleKeyDown.bind(this);
    this.handleWinResize = this.handleWinResize.bind(this);
    this.handleBackgroundClick = this.handleBackgroundClick.bind(this);
  }

  show() {
    this.el = document.createElement('div');
    this.el.classList.add('lightbox');
    this.el.innerHTML = '<a href="javascript:;" class="lightbox__close"></a>' +
      '<a class="lightbox__image lightbox__image-left"><img class="lightbox__img"></a>' +
      '<a class="lightbox__image lightbox__image-current"><img class="lightbox__img"></a>' +
      '<a class="lightbox__image lightbox__image-right"><img class="lightbox__img"></a>';

    document.body.appendChild(this.el);
    document.body.classList.add('lightbox-open');

    document.addEventListener('keydown', this.handleKeyDown);
    window.addEventListener('resize', this.handleWinResize);

    this.el.addEventListener('click', this.handleBackgroundClick);
    this.el.querySelector('.lightbox__close').addEventListener('click', () => this.hide());

    this.bindEvents({
      'click .lightbox__image-left': (el, e) => this.handleImageClick(e, -1),
      'click .lightbox__image-current': (el, e) => this.handleImageClick(e, 1),
      'click .lightbox__image-right': (el, e) => this.handleImageClick(e, 1),
    });

    this.updateImages();
  }

  hide() {
    document.body.removeChild(this.el);
    document.body.classList.remove('lightbox-open');
    document.removeEventListener('keydown', this.handleKeyDown);
    window.removeEventListener('resize', this.handleWinResize);
  }

  handleKeyDown(e) {
    if (e.keyCode == 37) this.move(-1);
    if (e.keyCode == 39) this.move(1);
    if (e.keyCode == 27) this.hide();
  }

  handleWinResize() {
    this.updateImages({ repositionOnly: true });
  }

  handleBackgroundClick(e) {
    if (e.target != this.el) return;
    this.hide();
  }

  handleImageClick(e, direction) {
    e.preventDefault();
    this.move(direction);
  }

  move(amount) {
    let newIndex = Math.max(Math.min(this.index + amount, this.images.length - 1), 0);
    if (newIndex == this.index) return;
    this.index = newIndex;
    this.updateImages();
  }

  updateImages(opts = { repositionOnly: false }) {
    const left = this.prepareImage(this.index - 1);
    const current = this.prepareImage(this.index);
    const right = this.prepareImage(this.index + 1);
    this.updateImage(left, this.el.querySelector('.lightbox__image-left'), opts);
    this.updateImage(current, this.el.querySelector('.lightbox__image-current'), opts);
    this.updateImage(right, this.el.querySelector('.lightbox__image-right'), opts);
    if (!opts.repositionOnly) {
      this.updatePreload(current, [right, left].filter(x => x != null));
    }
  }

  updateImage(data, el, opts) {
    const imgEl = el.querySelector('.lightbox__img');
    el.dataset.index = data ? data.index : -1;
    el.style.height = data ? data.height + 'px' : '';
    el.href = data ? data.src : '';
    imgEl.style.display = data ? 'block' : 'none';
    imgEl.style.width = data ? data.width + 'px' : '';
    imgEl.style.height = data ? data.height + 'px' : '';
    if (!opts.repositionOnly) {
      imgEl.src = data ? data.thumbSrc : '';
    }
  }

  prepareImage(index) {
    const def = this.images[index];
    if (!def) return null;

    const availableSpace = {
      width: Math.floor(document.documentElement.clientWidth * 0.84),
      height: document.documentElement.clientHeight - 80
    };

    const finalSize = this.fit(def, availableSpace);

    return { ...def, ...finalSize };
  }

  fit(size, within) {
    const widthScale = within.width / size.width;
    const heightScale = within.height / size.height;
    let scale = widthScale < heightScale ? widthScale : heightScale;
    if (scale > 1) scale = 1;
    return { width: Math.round(size.width * scale), height: Math.round(size.height * scale) };
  }

  updatePreload(active, next) {
    this.preloadQueue = next;
    this.startPreload(active);
  }

  startPreload(active) {
    if (this.currentPreload) {
      // already loading the same image, we can skip this
      if (this.currentPreload.def.index == active.index) return;

      // cancel previous preload
      this.currentPreload.img.src = '';
    }

    var img = new Image();
    var preload = { def: active, img };
    img.onload = () => {
      this.preloadFinished(preload);
    };

    this.currentPreload = preload;

    img.src = active.src;

    if (img.complete) {
      img.onload();
      img.onload
    }
  }

  preloadFinished(preload) {
    var imgEl = find(
      this.el.querySelectorAll('.lightbox__img'),
      x => x.parentNode.dataset.index == preload.def.index
    );

    if (imgEl) {
      imgEl.src = preload.def.src;
    }

    if (this.currentPreload != preload) {
      // this isn't actually the active preload, so it isn't our job to fire off the next one
      // (this state shouldn't happen, but could if the browser calls onload after we cancel a preload)
      return;
    }

    this.currentPreload = null;

    if (this.preloadQueue.length > 0) {
      this.startPreload(this.preloadQueue.shift());
    }
  }
}

document.addEventListener('DOMContentLoaded', () => {
  each(document.querySelectorAll('.js-columnize'), el => {
    columnize(el, Number(el.dataset.columns || 2))
  });

  each(document.querySelectorAll('.tag-summary'), el => TagSummary.attach(el));
  each(document.querySelectorAll('.dropdown'), el => Dropdown.attach(el));
  each(document.querySelectorAll('.gallery'), el => Gallery.attach(el));

  each(document.querySelectorAll('.navbar__toggler'), el => {
    el.addEventListener('click', () => {
      el.closest('.navbar').classList.toggle('navbar--expanded');
    });
  })
});
